#ifndef TYPON_CORE_FUTURE_HPP_INCLUDED
#define TYPON_CORE_FUTURE_HPP_INCLUDED

#include <atomic>
#include <cstdint>
#include <memory>
#include <type_traits>

#include <typon/fundamental/scope.hpp>

#include <typon/core/deque.hpp>
#include <typon/core/scheduler.hpp>


namespace typon
{

  template <typename T>
  struct Future
  {
    static constexpr std::uintptr_t ready {0};
    static constexpr std::uintptr_t no_waiter {1};

    std::atomic_uintptr_t _state {no_waiter};
    bool _consumed {false};
    union
    {
      T _value;
    };

    ~Future()
    {
      if (!_consumed)
      {
        if (_state.load(std::memory_order_acquire) == ready)
        {
          std::destroy_at(std::addressof(_value));
        }
      }
    }

    void put(T value)
    {
      std::construct_at(std::addressof(_value), std::move(value));
      auto state = _state.exchange(ready, std::memory_order_acq_rel);
      if (state != no_waiter)
      {
        Scheduler::enable(reinterpret_cast<Deque *>(state));
      }
    }

    bool await_ready() noexcept
    {
      return _state.load(std::memory_order_acquire) == ready;
    }

    void await_suspend(std::coroutine_handle<> coroutine) noexcept
    {
      auto deque = Scheduler::suspend(coroutine);
      auto state = reinterpret_cast<std::uintptr_t>(deque);
      if (_state.exchange(state, std::memory_order_acq_rel) == ready)
      {
        Scheduler::enable(deque);
      }
    }

    T await_resume() noexcept
    {
      fdt::defer defer( [this]() { std::destroy_at(std::addressof(_value)); });
      _consumed = true;
      return std::move(_value);
    }
  };


  template <typename T>
    requires std::is_trivially_destructible_v<T>
  struct Future<T>
  {
    static constexpr std::uintptr_t ready {0};
    static constexpr std::uintptr_t no_waiter {1};

    std::atomic_uintptr_t _state {no_waiter};
    union
    {
      T _value;
    };

    void put(T value)
    {
      std::construct_at(std::addressof(_value), std::move(value));
      auto state = _state.exchange(ready, std::memory_order_acq_rel);
      if (state != no_waiter)
      {
        Scheduler::enable(reinterpret_cast<Deque *>(state));
      }
    }

    bool await_ready() noexcept
    {
      return _state.load(std::memory_order_acquire) == ready;
    }

    void await_suspend(std::coroutine_handle<> coroutine) noexcept
    {
      auto deque = Scheduler::suspend(coroutine);
      auto state = reinterpret_cast<std::uintptr_t>(deque);
      if (_state.exchange(state, std::memory_order_acq_rel) == ready)
      {
        Scheduler::enable(deque);
      }
    }

    T await_resume() noexcept
    {
      return std::move(_value);
    }
  };


  template <typename T>
  struct Future<T&>
  {
    static constexpr std::uintptr_t ready {0};
    static constexpr std::uintptr_t no_waiter {1};

    std::atomic_uintptr_t _state {no_waiter};
    T* _value;

    void put(T& value)
    {
      _value = std::addressof(value);
      auto state = _state.exchange(ready, std::memory_order_acq_rel);
      if (state != no_waiter)
      {
        Scheduler::enable(reinterpret_cast<Deque *>(state));
      }
    }

    bool await_ready() noexcept
    {
      return _state.load(std::memory_order_acquire) == ready;
    }

    void await_suspend(std::coroutine_handle<> coroutine) noexcept
    {
      auto deque = Scheduler::suspend(coroutine);
      auto state = reinterpret_cast<std::uintptr_t>(deque);
      if (_state.exchange(state, std::memory_order_acq_rel) == ready)
      {
        Scheduler::enable(deque);
      }
    }

    T& await_resume() noexcept
    {
      return *_value;
    }
  };


  template <>
  struct Future<void>
  {
    static constexpr std::uintptr_t ready {0};
    static constexpr std::uintptr_t no_waiter {1};

    std::atomic_uintptr_t _state {no_waiter};

    void put()
    {
      auto state = _state.exchange(ready, std::memory_order_acq_rel);
      if (state != no_waiter)
      {
        Scheduler::enable(reinterpret_cast<Deque *>(state));
      }
    }

    bool await_ready() noexcept
    {
      return _state.load(std::memory_order_acquire) == ready;
    }

    void await_suspend(std::coroutine_handle<> coroutine) noexcept
    {
      auto deque = Scheduler::suspend(coroutine);
      auto state = reinterpret_cast<std::uintptr_t>(deque);
      if (_state.exchange(state, std::memory_order_acq_rel) == ready)
      {
        Scheduler::enable(deque);
      }
    }

    void await_resume() noexcept {}
  };

}


#endif // TYPON_CORE_FUTURE_HPP_INCLUDED
