#ifndef TYPON_CORE_WORKER_HPP_INCLUDED
#define TYPON_CORE_WORKER_HPP_INCLUDED

#include <atomic>
#include <coroutine>
#include <mutex>
#include <type_traits>
#include <utility>
#include <variant>
#include <vector>

#include <typon/fundamental/garbage_collector.hpp>
#include <typon/fundamental/random.hpp>

#include <typon/core/continuation.hpp>


namespace typon
{

  struct Worker
  {
    struct Work
    {
      static_assert(std::is_trivially_destructible_v<Continuation>);
      enum State : char { Empty, Resumable, Stolen };

      State _state;
      union
      {
        Deque * _deque;
        Continuation _task;
      };

      Work() noexcept : _state(Empty) {}

      Work(Deque * deque) noexcept : _state(Resumable), _deque(deque) {}

      Work(Continuation task) noexcept : _state(Stolen), _task(task) {}

      operator bool() noexcept
      {
        return _state != Empty;
      }
    };

    std::mutex _mutex;
    std::atomic<Deque *> _deque {nullptr};
    std::vector<Deque *> _pool;
    std::atomic_uint_fast64_t * _stealables;
    fdt::lock_free::garbage_collector * _gc;

    ~Worker()
    {
      for (auto & deque : _pool)
      {
        delete deque;
      }
      if (auto deque = _deque.load())
      {
        delete deque;
      }
    }

    void add(Deque * deque) noexcept
    {
      std::lock_guard lock(_mutex);
      _pool.push_back(deque);
    }

    bool try_add(Deque * deque) noexcept
    {
      if (!_mutex.try_lock())
      {
        return false;
      }
      std::lock_guard lock(_mutex, std::adopt_lock);
      _pool.push_back(deque);
      return true;
    }

    auto suspend(std::coroutine_handle<> coroutine) noexcept
    {
      auto deque = _deque.load();
      _deque.store(nullptr);
      deque->suspend(coroutine);
      return deque;
    }

    void resume(Work & work) noexcept
    {
      if (work._state == Work::Resumable)
      {
        auto deque = _deque.load();
        _deque.store(work._deque);
        if (deque)
        {
          _gc->retire(deque);
        }
        _stealables->fetch_add(1);
        work._deque->resume();
      }
      else
      {
        if (!_deque.load())
        {
          _deque.store(new Deque());
        }
        _stealables->fetch_add(1);
        work._task.resume();
      }
      if (_deque.load())
      {
        _stealables->fetch_sub(1);
      }
    }

    void push(Continuation task) noexcept
    {
      _deque.load()->push(task);
    }

    bool pop() noexcept
    {
      Deque * deque = _deque.load();
      bool result = deque->pop();
      if (auto garbage = deque->reclaim())
      {
        _gc->retire(garbage);
      }
      return result;
    }

    Work steal() noexcept
    {
      if (!_mutex.try_lock())
      {
        return {};
      }
      std::lock_guard lock(_mutex, std::adopt_lock);
      auto deque = _deque.load();
      auto total = _pool.size() + bool(deque);
      if (total == 0)
      {
        return {};
      }
      auto index = fdt::random::random64() % total;
      if (index == _pool.size())
      {
        if (auto task = deque->steal())
        {
          task.thefts()++;
          return task;
        }
        return {};
      }
      deque = _pool[index];
      if (deque->_state.load() == Deque::Resumable)
      {
        if (index < _pool.size() - 1)
        {
          _pool[index] = _pool.back();
        }
        _pool.pop_back();
        return deque;
      }
      auto task = deque->pop_top();
      if (auto garbage = deque->reclaim())
      {
        delete garbage;
      }
      if (task)
      {
        task.thefts()++;
        return task;
      }
      if (index < _pool.size() - 1)
      {
        _pool[index] = _pool.back();
      }
      _pool.pop_back();
      Deque::State expected = Deque::Suspended;
      if (!deque->_state.compare_exchange_strong(expected, Deque::Empty))
      {
        return deque;
      }
      _stealables->fetch_sub(1);
      return {};
    }
  };

}


#endif // TYPON_CORE_WORKER_HPP_INCLUDED
