#ifndef TYPON_CORE_SCHEDULER_HPP_INCLUDED
#define TYPON_CORE_SCHEDULER_HPP_INCLUDED

#include <atomic>
#include <coroutine>
#include <cstdint>
#include <exception>
#include <thread>
#include <vector>

#include <typon/fundamental/event_count.hpp>
#include <typon/fundamental/garbage_collector.hpp>
#include <typon/fundamental/random.hpp>

#include <typon/core/continuation.hpp>
#include <typon/core/deque.hpp>
#include <typon/core/worker.hpp>


namespace typon
{

  struct Scheduler
  {
    using uint = unsigned int;
    using u64 = std::uint_fast64_t;
    using Work = Worker::Work;
    using garbage_collector = fdt::lock_free::garbage_collector;

    static inline thread_local uint thread_id;

    static Scheduler & get() noexcept
    {
      static Scheduler scheduler {std::thread::hardware_concurrency()};
      return scheduler;
    }

    static void schedule(std::coroutine_handle<> task) noexcept
    {
      uint id = fdt::random::random() % get()._concurrency;
      get()._worker[id].add(new Deque(task));
      get()._stealables.fetch_add(1);
      get()._notifyer.notify_one();
    }

    static void push(Continuation task) noexcept
    {
      get()._worker[thread_id].push(task);
    }

    static bool pop() noexcept
    {
      return get()._worker[thread_id].pop();
    }

    static auto suspend(std::coroutine_handle<> coroutine) noexcept
    {
      Worker & worker = get()._worker[thread_id];
      auto deque = worker.suspend(coroutine);
      uint id = fdt::random::random() % get()._concurrency;
      get()._worker[id].add(deque);
      return deque;
    }

    static void enable(Deque * deque) noexcept
    {
      auto state = deque->_state.exchange(Deque::Resumable);
      if (state == Deque::Empty)
      {
        uint id = fdt::random::random() % get()._concurrency;
        get()._worker[id].add(deque);
        get()._stealables.fetch_add(1);
      }
      get()._notifyer.notify_one();
    }

    std::atomic<uint> _thieves = 0;
    std::atomic<u64> _stealables = 0;
    std::vector<Worker> _worker;
    std::vector<std::thread> _thread;
    std::atomic_bool _done {false};
    fdt::lock_free::event_count<> _notifyer;
    const uint _concurrency;
    garbage_collector _gc;

    Scheduler(uint concurrency) noexcept
      : _worker(concurrency)
      , _concurrency(concurrency)
      , _gc(concurrency)
    {
      for (uint i = 0; i < concurrency; i++)
      {
        _worker[i]._gc = &(_gc);
        _worker[i]._stealables = &(_stealables);
      }

      thread_id = concurrency;

      for (uint id = 0; id < concurrency; id++)
      {
        _thread.emplace_back([this, id]() {
          thread_id = id;

          Work work {};
          for(;;)
          {
            if (!wait_for_work(work))
            {
              break;
            }
            exploit_work(work);
          }
        });
      }
    }

    ~Scheduler() noexcept
    {
      _done.store(true);
      _notifyer.notify_all();
      for (auto & t : _thread)
      {
        t.join();
      }
    }

    void exploit_work(Work & work) noexcept
    {
      _worker[thread_id].resume(work);
    }

    void explore_work(Work & work) noexcept
    {
      auto epoch = _gc.epoch(thread_id);
      for (uint i = 0; i < _concurrency * 2 + 1; i++)
      {
        uint id = fdt::random::random() % _concurrency;
        work = _worker[id].steal();
        if (work)
        {
          break;
        }
      }
    }

    bool wait_for_work(Work & work) noexcept
    {
      work = {};
      for(;;)
      {
        _thieves.fetch_add(1);
        explore_work(work);
        if (work)
        {
          if (_thieves.fetch_sub(1) == 1)
          {
            _notifyer.notify_one();
          }
          return true;
        }
        auto key = _notifyer.prepare_wait();
        if (_done.load())
        {
          _notifyer.cancel_wait();
          _notifyer.notify_all();
          _thieves.fetch_sub(1);
          return false;
        }
        if (_thieves.fetch_sub(1) == 1)
        {
          if (_stealables.load() > 0)
          {
            _notifyer.cancel_wait();
            continue;
          }
        }
        _notifyer.wait(key);
      }
    }
  };

}

#endif // TYPON_SCHEDULER_HPP_INCLUDED
